/* Parallel Trapezoidal Rule
 *
 * Input: None.
 * Output:  Estimate of the integral from 0 to Pi/2 of
 *              f(x) = dx/(sqrt(1-(cos x * cos x)/2))
 *          using the trapezoidal rule and 128 trapezoids.
 *
 * Algorithm:
 *    1.  Each process calculates "its" interval of
 *        integration.
 *    2.  Each process estimates the integral of f(x)
 *        over its interval using the trapezoidal rule.
 *    3a. Each process != 0 sends its integral to 0.
 *    3b. Process 0 sums the calculations received from
 *        the individual processes and prints the result.
 *
 */

#include <stdio.h> 
#include <stdlib.h>
#include <math.h>

/* using MPI routines, definitions, etc. */
#include "mpi.h"

#define ping 101
#define pong 101

main(int argc, char** argv) {

    int         each_process_rank;   /* Each process rank */
    int         process_num;         /* The number of processes */

    int         process_source;    /* Process sending integral  */
    int         process_dest = 0;  /* All messages go to Process 0 */
	int         length;
	float       buffer[100000];
    double      start, finish, time;

    MPI_Status  status;

    /* Let the system do what it needs to start up MPI */
    MPI_Init(&argc, &argv);

    /* Get each process rank */
    MPI_Comm_rank(MPI_COMM_WORLD, &each_process_rank);

    /* Find out how many processes are being used */
    MPI_Comm_size(MPI_COMM_WORLD, &process_num);

    /* Add up the integrals calculated by each process */
    if (each_process_rank == 0) {

		printf("Length\tTotal Time\tTransfer Rate\n");

    	for (length = 1; length <=  100000; length += 1000){
        	start = MPI_Wtime();
 
        	for (i = 1; i <= 100; i++){
 
        		for (process_source = 1; process_source < process_num; 
					  process_source++) {
 
            		MPI_Send(buffer, length, MPI_FLOAT, process_source, ping,
                                 MPI_COMM_WORLD);
				}

				for (process_source = 1; process_source < process_num;
					  process_source++) {

            		MPI_Recv(buffer, length, MPI_FLOAT, process_source, pong,
                    			 MPI_COMM_WORLD, &status);
        		}
		 	}

        	finish = MPI_Wtime();
 
        	time = finish - start;
 
        	printf("%d\t%f\t%f\n", length, time/200.,
               (float)(2 * 8 * 100 * length)/time);
 
    	}

    } else {  

    		for (length = 1; length <= 100000; length += 1000) {
 
            	for (i = 1; i <= 100; i++) {
 
                	MPI_Recv(buffer, length, MPI_FLOAT, process_dest, ping,
                                             MPI_COMM_WORLD, &status);
 
                	MPI_Send(buffer, length, MPI_FLOAT, process_dest, pong,
                                            MPI_COMM_WORLD);
 
            	}
    		}
    }

    MPI_Finalize();
} /*  main  */

